/**********************************************************************
Copyright (c) 2009 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.datanucleus.store.odf.fieldmanager;

import java.lang.reflect.Array;
import java.util.Calendar;
import java.util.Collection;
import java.util.Currency;
import java.util.Date;
import java.util.Map;

import javax.xml.datatype.XMLGregorianCalendar;

import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.ObjectManager;
import org.datanucleus.StateManager;
import org.datanucleus.exceptions.NucleusDataStoreException;
import org.datanucleus.exceptions.NucleusException;
import org.datanucleus.metadata.AbstractClassMetaData;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.ColumnMetaData;
import org.datanucleus.metadata.MetaDataUtils;
import org.datanucleus.metadata.Relation;
import org.datanucleus.sco.SCOUtils;
import org.datanucleus.store.fieldmanager.AbstractFieldManager;
import org.datanucleus.store.odf.ODFUtils;
import org.datanucleus.store.types.ObjectLongConverter;
import org.datanucleus.store.types.ObjectStringConverter;
import org.datanucleus.util.NucleusLogger;
import org.openoffice.odf.doc.element.table.OdfTableCell;
import org.openoffice.odf.doc.element.table.OdfTableRow;
import org.w3c.dom.NodeList;

/**
 * FieldManager for the fetch of fields from ODF.
 */
public class FetchFieldManager extends AbstractFieldManager
{
    protected final StateManager sm;
    protected final OdfTableRow row;

    public FetchFieldManager(StateManager sm, OdfTableRow row)
    {
        this.sm = sm;
        this.row = row;
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchBooleanField(int)
     */
    @Override
    public boolean fetchBooleanField(int fieldNumber)
    {
        int index = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        NodeList cellNodes = row.getChildNodes();
        OdfTableCell cell = (OdfTableCell)cellNodes.item(index);
        return cell.getBooleanValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchByteField(int)
     */
    @Override
    public byte fetchByteField(int fieldNumber)
    {
        int index = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        NodeList cellNodes = row.getChildNodes();
        OdfTableCell cell = (OdfTableCell)cellNodes.item(index);
        return cell.getValue().byteValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchCharField(int)
     */
    @Override
    public char fetchCharField(int fieldNumber)
    {
        int index = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        NodeList cellNodes = row.getChildNodes();
        OdfTableCell cell = (OdfTableCell)cellNodes.item(index);
        return cell.getStringValue().charAt(0);
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchDoubleField(int)
     */
    @Override
    public double fetchDoubleField(int fieldNumber)
    {
        int index = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        NodeList cellNodes = row.getChildNodes();
        OdfTableCell cell = (OdfTableCell)cellNodes.item(index);
        return cell.getValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchFloatField(int)
     */
    @Override
    public float fetchFloatField(int fieldNumber)
    {
        int index = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        NodeList cellNodes = row.getChildNodes();
        OdfTableCell cell = (OdfTableCell)cellNodes.item(index);
        return cell.getValue().floatValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchIntField(int)
     */
    @Override
    public int fetchIntField(int fieldNumber)
    {
        int index = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        NodeList cellNodes = row.getChildNodes();
        OdfTableCell cell = (OdfTableCell)cellNodes.item(index);
        return cell.getValue().intValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchLongField(int)
     */
    @Override
    public long fetchLongField(int fieldNumber)
    {
        int index = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        NodeList cellNodes = row.getChildNodes();
        OdfTableCell cell = (OdfTableCell)cellNodes.item(index);
        return cell.getValue().longValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchShortField(int)
     */
    @Override
    public short fetchShortField(int fieldNumber)
    {
        int index = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        NodeList cellNodes = row.getChildNodes();
        OdfTableCell cell = (OdfTableCell)cellNodes.item(index);
        return cell.getValue().shortValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchStringField(int)
     */
    @Override
    public String fetchStringField(int fieldNumber)
    {
        int index = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        NodeList cellNodes = row.getChildNodes();
        OdfTableCell cell = (OdfTableCell)cellNodes.item(index);
        String strVal = cell.getStringValue();
        if (strVal.equals(ODFUtils.STRING_NULL_VALUE))
        {
            return null;
        }
        return cell.getStringValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchObjectField(int)
     */
    @Override
    public Object fetchObjectField(int fieldNumber)
    {
        int index = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        NodeList cellNodes = row.getChildNodes();
        OdfTableCell cell = (OdfTableCell)cellNodes.item(index);

        ObjectManager om = sm.getObjectManager();
        ClassLoaderResolver clr = om.getClassLoaderResolver();
        AbstractMemberMetaData mmd =
            sm.getClassMetaData().getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        int relationType = mmd.getRelationType(clr);
        if (relationType == Relation.NONE)
        {
            Class type = mmd.getType();
            Object value = null;
            if (type == Boolean.class)
            {
                if (cell.getBooleanValue() == null)
                {
                    return null;
                }
                value = cell.getBooleanValue();
            }
            else if (type == Byte.class)
            {
                if (getCellDoubleValueAllowingForNull(cell) == null)
                {
                    return null;
                }
                value = cell.getValue().byteValue();
            }
            else if (type == Character.class)
            {
                if (cell.getStringValue() == null)
                {
                    return null;
                }
                value = cell.getStringValue().charAt(0);
            }
            else if (type == Double.class)
            {
                if (getCellDoubleValueAllowingForNull(cell) == null)
                {
                    return null;
                }
                value = cell.getValue();
            }
            else if (type == Float.class)
            {
                if (getCellDoubleValueAllowingForNull(cell) == null)
                {
                    return null;
                }
                value = cell.getValue();
            }
            else if (type == Integer.class)
            {
                if (getCellDoubleValueAllowingForNull(cell) == null)
                {
                    return null;
                }
                value = cell.getValue().intValue();
            }
            else if (type == Long.class)
            {
                if (getCellDoubleValueAllowingForNull(cell) == null)
                {
                    return null;
                }
                value = cell.getValue().longValue();
            }
            else if (type == Short.class)
            {
                if (getCellDoubleValueAllowingForNull(cell) == null)
                {
                    return null;
                }
                value = cell.getValue().shortValue();
            }
            else if (Calendar.class.isAssignableFrom(type))
            {
                if (cell.getDateValue() == null)
                {
                    return null;
                }
                XMLGregorianCalendar xmlCal = cell.getDateValue();
                value = xmlCal.toGregorianCalendar();
            }
            else if (Date.class.isAssignableFrom(type))
            {
                if (cell.getDateValue() == null)
                {
                    return null;
                }
                XMLGregorianCalendar xmlCal = cell.getDateValue();
                Date date = xmlCal.toGregorianCalendar().getTime();
                if (type == java.sql.Date.class)
                {
                    value = new java.sql.Date(date.getTime());
                }
                else if (type == java.sql.Time.class)
                {
                    value = new java.sql.Time(date.getTime());
                }
                else if (type == java.sql.Timestamp.class)
                {
                    value = new java.sql.Timestamp(date.getTime());
                }
                else
                {
                    value = date;
                }
            }
            else if (Currency.class.isAssignableFrom(type))
            {
                if (cell.getCurrency() == null)
                {
                    return null;
                }
                value = cell.getCurrency();
            }
            else if (Enum.class.isAssignableFrom(type))
            {
                String cellValue = cell.getStringValue();
                if (cellValue != null && cellValue.length() > 0)
                {
                    value = Enum.valueOf(type, cell.getStringValue());
                }
                else
                {
                    value = null;
                }
            }
            else
            {
                boolean useLong = false;
                ColumnMetaData[] colmds = mmd.getColumnMetaData();
                if (colmds != null && colmds.length == 1)
                {
                    String jdbc = colmds[0].getJdbcType();
                    if (jdbc != null && (jdbc.equalsIgnoreCase("int") || jdbc.equalsIgnoreCase("integer")))
                    {
                        useLong = true;
                    }
                }
                ObjectStringConverter strConv = 
                    sm.getObjectManager().getOMFContext().getTypeManager().getStringConverter(mmd.getType());
                ObjectLongConverter longConv = 
                    sm.getObjectManager().getOMFContext().getTypeManager().getLongConverter(mmd.getType());
                if (useLong && longConv != null)
                {
                    value = longConv.toObject(cell.getValue().longValue());
                }
                else if (!useLong && strConv != null)
                {
                    String cellValue = cell.getStringValue();
                    if (cellValue != null && cellValue.length() > 0)
                    {
                        value = strConv.toObject(cell.getStringValue());
                    }
                    else
                    {
                        return null;
                    }
                }
                else if (!useLong && longConv != null)
                {
                    value = longConv.toObject(cell.getValue().longValue());
                }
                else
                {
                    // Not supported as String so just set to null
                    NucleusLogger.PERSISTENCE.warn("Field " + mmd.getFullFieldName() + 
                        " could not be set in the object since it is not persistable to ODF");
                    return null;
                }
            }
            return sm.wrapSCOField(fieldNumber, value, false, false, true);
        }
        else if (relationType == Relation.ONE_TO_ONE_BI || relationType == Relation.ONE_TO_ONE_UNI ||
            relationType == Relation.MANY_TO_ONE_BI)
        {
            // Persistable object - retrieve the string form of the identity, and find the object
            String idStr = cell.getStringValue();
            if (idStr == null)
            {
                return null;
            }

            if (idStr.startsWith("[") && idStr.endsWith("]"))
            {
                idStr = idStr.substring(1, idStr.length()-1);
                AbstractClassMetaData relatedCmd = om.getMetaDataManager().getMetaDataForClass(mmd.getType(), clr);
                return getObjectFromIdString(idStr, relatedCmd);
            }
            else
            {
                return null;
            }
        }
        else if (relationType == Relation.MANY_TO_MANY_BI || relationType == Relation.ONE_TO_MANY_BI ||
            relationType == Relation.ONE_TO_MANY_UNI)
        {
            // Collection/Map/Array
            String cellStr = cell.getStringValue();
            if (cellStr == null)
            {
                return null;
            }

            if (cellStr.startsWith("[") && cellStr.endsWith("]"))
            {
                cellStr = cellStr.substring(1, cellStr.length()-1);
                String[] components = MetaDataUtils.getInstance().getValuesForCommaSeparatedAttribute(cellStr);
                if (Collection.class.isAssignableFrom(mmd.getType()))
                {
                    Collection<Object> coll;
                    try
                    {
                        Class instanceType = SCOUtils.getContainerInstanceType(mmd.getType(), mmd.getOrderMetaData() != null);
                        coll = (Collection<Object>) instanceType.newInstance();
                    }
                    catch (Exception e)
                    {
                        throw new NucleusDataStoreException(e.getMessage(), e);
                    }

                    if (components != null)
                    {
                        for (int i=0;i<components.length;i++)
                        {
                            AbstractClassMetaData elementCmd = mmd.getCollection().getElementClassMetaData(
                                om.getClassLoaderResolver(), om.getMetaDataManager());
                            Object element = getObjectFromIdString(components[i], elementCmd);
                            coll.add(element);
                        }
                    }
                    return sm.wrapSCOField(fieldNumber, coll, false, false, true);
                }
                else if (Map.class.isAssignableFrom(mmd.getType()))
                {
                    // TODO Implement map retrieval
                }
                else if (mmd.getType().isArray())
                {
                    Object array = null;
                    if (components != null)
                    {
                        array = Array.newInstance(mmd.getType().getComponentType(), components.length);
                        for (int i=0;i<components.length;i++)
                        {
                            AbstractClassMetaData elementCmd = mmd.getCollection().getElementClassMetaData(
                                om.getClassLoaderResolver(), om.getMetaDataManager());
                            Object element = getObjectFromIdString(components[i], elementCmd);
                            Array.set(array, i, element);
                        }
                    }
                    else
                    {
                        array = Array.newInstance(mmd.getType().getComponentType(), 0);
                    }
                    return sm.wrapSCOField(fieldNumber, array, false, false, true);
                }
            }
            throw new NucleusException("Dont currently support retrieval of collection/map/array types from ODF");
        }
        else
        {
            throw new NucleusException("Dont currently support retrieval of type " + mmd.getTypeName() + " from ODF");
        }
    }

    private Double getCellDoubleValueAllowingForNull(OdfTableCell cell)
    {
        try
        {
            return cell.getValue();
        }
        catch (NumberFormatException nfe)
        {
            return null;
        }
    }

    /**
     * Convenience method to find an object given a string form of its identity, and the metadata for the
     * class (or a superclass).
     * @param idStr The id string
     * @param cmd Metadata for the class
     * @return The object
     */
    protected Object getObjectFromIdString(String idStr, AbstractClassMetaData cmd)
    {
        ObjectManager om = sm.getObjectManager();
        ClassLoaderResolver clr = om.getClassLoaderResolver();
        Object id = null;
        if (cmd.usesSingleFieldIdentityClass())
        {
            id = om.getApiAdapter().getNewApplicationIdentityObjectId(clr, cmd, idStr);
        }
        else
        {
            Class cls = clr.classForName(cmd.getFullClassName());
            id = om.newObjectId(cls, idStr);
        }
        return om.findObject(id, true, true, null);
    }
}