/**********************************************************************
Copyright (c) 2009 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors :
 ...
***********************************************************************/
package org.datanucleus.store.odf.fieldmanager;

import java.lang.reflect.Array;
import java.util.Calendar;
import java.util.Collection;
import java.util.Currency;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Iterator;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;

import org.datanucleus.StateManager;
import org.datanucleus.exceptions.NucleusException;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.ColumnMetaData;
import org.datanucleus.metadata.IdentityType;
import org.datanucleus.metadata.Relation;
import org.datanucleus.store.fieldmanager.AbstractFieldManager;
import org.datanucleus.store.odf.ODFUtils;
import org.datanucleus.store.types.ObjectLongConverter;
import org.datanucleus.store.types.ObjectStringConverter;
import org.datanucleus.util.NucleusLogger;
import org.openoffice.odf.doc.OdfFileDom;
import org.openoffice.odf.doc.element.table.OdfTableCell;
import org.openoffice.odf.doc.element.table.OdfTableRow;
import org.openoffice.odf.dom.type.office.OdfValueType;

/**
 * FieldManager to handle the store information into an ODF spreadsheet row using an object.
 */
public class InsertFieldManager extends AbstractFieldManager
{
    /** StateManager for object being inserted. */
    private final StateManager sm;

    /** The cells, populated during calls to storeXXX. */
    private final OdfTableCell[] cells;

    public InsertFieldManager(StateManager sm, OdfFileDom contentDoc, OdfTableRow row)
    {
        this.sm = sm;

        // Create all cells in row so we can use user positions during population
        int numCells = sm.getClassMetaData().getAllMemberPositions().length;
        if (sm.getClassMetaData().getIdentityType() == IdentityType.DATASTORE)
        {
            numCells++;
        }
        if (sm.getClassMetaData().hasVersionStrategy())
        {
            numCells++;
        }
        cells = new OdfTableCell[numCells];
        for (int i=0;i<numCells;i++)
        {
            cells[i] = new OdfTableCell(contentDoc);
            row.appendCell(cells[i]);
        }
    }

    public void storeBooleanField(int fieldNumber, boolean value)
    {
        OdfTableCell cell = cells[ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber)];
        cell.setValueType(OdfValueType.BOOLEAN);
        cell.setBooleanValue(value);
        AbstractMemberMetaData mmd = 
            sm.getClassMetaData().getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        if (mmd.isPrimaryKey())
        {
            cell.setStyleName("DN_PK");
        }
    }

    public void storeByteField(int fieldNumber, byte value)
    {
        OdfTableCell cell = cells[ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber)];
        cell.setValueType(OdfValueType.FLOAT);
        cell.setValue(new Double(value));
        AbstractMemberMetaData mmd = 
            sm.getClassMetaData().getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        if (mmd.isPrimaryKey())
        {
            cell.setStyleName("DN_PK");
        }
    }

    public void storeCharField(int fieldNumber, char value)
    {
        OdfTableCell cell = cells[ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber)];
        cell.setValueType(OdfValueType.STRING);
        cell.setStringValue("" + value);
        AbstractMemberMetaData mmd = 
            sm.getClassMetaData().getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        if (mmd.isPrimaryKey())
        {
            cell.setStyleName("DN_PK");
        }
    }

    public void storeDoubleField(int fieldNumber, double value)
    {
        OdfTableCell cell = cells[ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber)];
        cell.setValueType(OdfValueType.FLOAT);
        cell.setValue(value);
        AbstractMemberMetaData mmd = 
            sm.getClassMetaData().getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        if (mmd.isPrimaryKey())
        {
            cell.setStyleName("DN_PK");
        }
    }

    public void storeFloatField(int fieldNumber, float value)
    {
        OdfTableCell cell = cells[ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber)];
        cell.setValueType(OdfValueType.FLOAT);
        cell.setValue(new Double(value));
        AbstractMemberMetaData mmd = 
            sm.getClassMetaData().getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        if (mmd.isPrimaryKey())
        {
            cell.setStyleName("DN_PK");
        }
    }

    public void storeIntField(int fieldNumber, int value)
    {
        OdfTableCell cell = cells[ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber)];
        cell.setValueType(OdfValueType.FLOAT);
        cell.setValue(new Double(value));
        AbstractMemberMetaData mmd = 
            sm.getClassMetaData().getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        if (mmd.isPrimaryKey())
        {
            cell.setStyleName("DN_PK");
        }
    }

    public void storeLongField(int fieldNumber, long value)
    {
        OdfTableCell cell = cells[ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber)];
        cell.setValueType(OdfValueType.FLOAT);
        cell.setValue(new Double(value));
        AbstractMemberMetaData mmd = 
            sm.getClassMetaData().getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        if (mmd.isPrimaryKey())
        {
            cell.setStyleName("DN_PK");
        }
    }

    public void storeShortField(int fieldNumber, short value)
    {
        OdfTableCell cell = cells[ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber)];
        cell.setValueType(OdfValueType.FLOAT);
        cell.setValue(new Double(value));
        AbstractMemberMetaData mmd = 
            sm.getClassMetaData().getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        if (mmd.isPrimaryKey())
        {
            cell.setStyleName("DN_PK");
        }
    }

    public void storeStringField(int fieldNumber, String value)
    {
        OdfTableCell cell = cells[ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber)];
        cell.setValueType(OdfValueType.STRING);
        if (value == null)
        {
            cell.setStringValue(ODFUtils.STRING_NULL_VALUE);
        }
        else
        {
            cell.setStringValue(value);
        }
        AbstractMemberMetaData mmd = 
            sm.getClassMetaData().getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        if (mmd.isPrimaryKey())
        {
            cell.setStyleName("DN_PK");
        }
    }

    public void storeObjectField(int fieldNumber, Object value)
    {
        AbstractMemberMetaData mmd = 
            sm.getClassMetaData().getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        int relationType = mmd.getRelationType(sm.getObjectManager().getClassLoaderResolver());

        OdfTableCell cell = cells[ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber)];
        if (mmd.isPrimaryKey())
        {
            cell.setStyleName("DN_PK");
        }
        else if (relationType != Relation.NONE)
        {
            cell.setStyleName("DN_Relation");
        }

        if (value == null)
        {
            if (Number.class.isAssignableFrom(mmd.getType()))
            {
                cell.setValueType(OdfValueType.FLOAT);
            }
            else if (Date.class.isAssignableFrom(mmd.getType()) || Calendar.class.isAssignableFrom(mmd.getType()))
            {
                cell.setValueType(OdfValueType.DATE);
            }
            else
            {
                // Assume it's a String type TODO Set the type based on the field type
                cell.setValueType(OdfValueType.STRING);
            }
        }
        else if (relationType == Relation.NONE)
        {
            if (value instanceof Calendar)
            {
                cell.setValueType(OdfValueType.DATE);
                try
                {
                    cell.setDateValue(DatatypeFactory.newInstance().newXMLGregorianCalendar((GregorianCalendar)value));
                }
                catch (DatatypeConfigurationException e)
                {
                    throw new NucleusException("Error occurred inserting field with value=" + value, e);
                }
            }
            else if (value instanceof Date)
            {
                cell.setValueType(OdfValueType.DATE);
                try
                {
                    GregorianCalendar cal = (GregorianCalendar)GregorianCalendar.getInstance();
                    cal.setTime((Date)value);
                    cell.setDateValue(DatatypeFactory.newInstance().newXMLGregorianCalendar(cal));
                }
                catch (DatatypeConfigurationException e)
                {
                    throw new NucleusException("Error occurred inserting field with value=" + value, e);
                }
            }
            else if (value instanceof Boolean)
            {
                cell.setValueType(OdfValueType.BOOLEAN);
                cell.setBooleanValue((Boolean)value);
            }
            else if (value instanceof Byte)
            {
                cell.setValueType(OdfValueType.FLOAT);
                cell.setValue(new Double((Byte)value));
            }
            else if (value instanceof Character)
            {
                cell.setValueType(OdfValueType.STRING);
                cell.setStringValue("" + value);
            }
            else if (value instanceof Double)
            {
                cell.setValueType(OdfValueType.FLOAT);
                cell.setValue((Double)value);
            }
            else if (value instanceof Float)
            {
                cell.setValueType(OdfValueType.FLOAT);
                cell.setValue((Double)value);
            }
            else if (value instanceof Integer)
            {
                cell.setValueType(OdfValueType.FLOAT);
                cell.setValue(new Double((Integer)value));
            }
            else if (value instanceof Long)
            {
                cell.setValueType(OdfValueType.FLOAT);
                cell.setValue(new Double((Long)value));
            }
            else if (value instanceof Short)
            {
                cell.setValueType(OdfValueType.FLOAT);
                cell.setValue(new Double((Short)value));
            }
            else if (value instanceof Currency)
            {
                ObjectStringConverter converter = 
                    sm.getObjectManager().getOMFContext().getTypeManager().getStringConverter(value.getClass());
                cell.setValueType(OdfValueType.CURRENCY);
                cell.setCurrency(converter.toString(value));
            }
            else if (value instanceof Enum)
            {
                cell.setValueType(OdfValueType.STRING);
                cell.setStringValue(((Enum)value).name());
            }
            else
            {
                // See if we can persist it as a String
                boolean useLong = false;
                ColumnMetaData[] colmds = mmd.getColumnMetaData();
                if (colmds != null && colmds.length == 1)
                {
                    String jdbc = colmds[0].getJdbcType();
                    if (jdbc != null && (jdbc.equalsIgnoreCase("int") || jdbc.equalsIgnoreCase("integer")))
                    {
                        useLong = true;
                    }
                }

                ObjectStringConverter strConv = 
                    sm.getObjectManager().getOMFContext().getTypeManager().getStringConverter(value.getClass());
                ObjectLongConverter longConv =
                    sm.getObjectManager().getOMFContext().getTypeManager().getLongConverter(value.getClass());
                if (useLong)
                {
                    if (longConv != null)
                    {
                        cell.setValueType(OdfValueType.FLOAT);
                        cell.setValue(Double.valueOf(longConv.toLong(value)));
                        return;
                    }
                }
                else
                {
                    if (strConv != null)
                    {
                        cell.setValueType(OdfValueType.STRING);
                        cell.setStringValue(strConv.toString(value));
                        return;
                    }
                    else if (longConv != null)
                    {
                        cell.setValueType(OdfValueType.FLOAT);
                        cell.setValue(Double.valueOf(longConv.toLong(value)));
                        return;
                    }
                }

                NucleusLogger.PERSISTENCE.warn("Dont currently support persistence of field=" + mmd.getFullFieldName() +
                    " type=" + value.getClass().getName() + " to ODF");
            }
        }
        else if (relationType == Relation.ONE_TO_ONE_BI || relationType == Relation.ONE_TO_ONE_UNI ||
            relationType == Relation.MANY_TO_ONE_BI)
        {
            // Persistable object - persist the related object and store the identity in the cell
            Object valuePC = sm.getObjectManager().persistObjectInternal(value, null, sm, fieldNumber, -1);
            Object valueId = sm.getObjectManager().getApiAdapter().getIdForObject(valuePC);
            cell.setValueType(OdfValueType.STRING);
            cell.setStringValue("[" + valueId.toString() + "]");
        }
        else if (relationType == Relation.MANY_TO_MANY_BI || relationType == Relation.ONE_TO_MANY_BI ||
            relationType == Relation.ONE_TO_MANY_UNI)
        {
            // Collection/Map/Array
            if (mmd.hasCollection())
            {
                StringBuffer cellValue = new StringBuffer("[");
                Collection coll = (Collection)value;
                Iterator collIter = coll.iterator();
                while (collIter.hasNext())
                {
                    Object element = collIter.next();
                    Object elementPC = sm.getObjectManager().persistObjectInternal(element, null, sm, fieldNumber, -1);
                    Object elementID = sm.getObjectManager().getApiAdapter().getIdForObject(elementPC);
                    cellValue.append(elementID.toString());
                    if (collIter.hasNext())
                    {
                        cellValue.append(",");
                    }
                }
                cellValue.append("]");
                cell.setValueType(OdfValueType.STRING);
                cell.setStringValue(cellValue.toString());
            }
            else if (mmd.hasMap())
            {
                // TODO Implement map persistence - what to do if key or value is non-PC
                throw new NucleusException("Dont currently support persistence of map types to ODF");
            }
            else if (mmd.hasArray())
            {
                StringBuffer cellValue = new StringBuffer("[");
                for (int i=0;i<Array.getLength(value);i++)
                {
                    Object element = Array.get(value, i);
                    Object elementPC = sm.getObjectManager().persistObjectInternal(element, null, sm, fieldNumber, -1);
                    Object elementID = sm.getObjectManager().getApiAdapter().getIdForObject(elementPC);
                    cellValue.append(elementID.toString());
                    if (i < (Array.getLength(value)-1))
                    {
                        cellValue.append(",");
                    }
                }
                cellValue.append("]");
                cell.setValueType(OdfValueType.STRING);
                cell.setStringValue(cellValue.toString());
            }
        }
    }
}